import numpy as np
import quaternion

from skeletal_animation.core.math import *
from skeletal_animation.core.model import *
from skeletal_animation.core.animate import *
from skeletal_animation.utils.loader import *
from skeletal_animation.algo.motionblending import *

"""
This program demonstrates the creation of a new skeletal animation
based on existing ones.

The starting animations are a walk and a death animation.
The goal is to create a new animation of a character walking and
then suddenly dying. The resulting animation through the uss of
blending, cropping, merging and translating existing animations.
"""

# Set up variables with the path to where our animations are stored
DATA_DIR = "D:/He-Arc/TB/tb-animation-squelettale/skeletal_animation/animated_models/"
# Folder for FBX files
FBX_DATA_DIR = DATA_DIR + "fbx/"
# Folder for serialized .pkl files
SERIALIZED_DATA_DIR = DATA_DIR + "serialized/"

# Load the walk animation from a FBX file
walk_animator = load_fbx(FBX_DATA_DIR + "walk.fbx")
# Load the death animation from a FBX file
death_animator = load_fbx(FBX_DATA_DIR + "death.fbx")

# We crop the walk animation into 2 halfs, one to be played at the beginning
# of the new animation and the other to be blended with the death animation
walk_start = walk_animator.crop(0.0, 0.5)
walk_end_for_blend = walk_animator.crop(0.5, 1.0)

# Similarly, we crop the death animation into 2 parts. The first part is
# blended with the walk animation and the second part is played at the end of
# the new animation
death_start_for_blend = death_animator.crop(0.0, 0.2)
death_end = death_animator.crop(0.2, 1.0)

# We define a blend curve for the blending between the walk and death animation
# The curve starts slowly and speeds up over time
blend_curve = Curve([Key(0.0, 0.0), Key(0.4, 0.5), Key(0.7, 0.9), Key(1.0, 1.0)])

# Before we blend the animations, we need to align them. We do this by translating the death
# animation to be aligned with the end of the first half of the walk animation.
# Here is how we do it:

# We get the position of the root bone at the end of the first half of the walk animation
walk_last_root_position = walk_start.animation.joint_animations[0].translation_curve.keys[-1].value
print(walk_last_root_position) # [4.16174078  102.57474518  96.60121155]

# We get the position of the root bone at the beginning of the death animation
death_first_root_position = death_start_for_blend.animation.joint_animations[0].translation_curve.keys[0].value
print(death_first_root_position) # [-0.40051448  101.26885986   1.7320683]

# We compute the translation vector that we need to apply to the death animation
translation_vector = walk_last_root_position - death_first_root_position
print(translation_vector) # [ 4.56225526  1.30588532 94.86914325]

# We translate both parts of the death animation
death_start_for_blend.translate(
    translation_vector[0],
    translation_vector[1],
    translation_vector[2]
    )

death_end.translate(
    translation_vector[0],
    translation_vector[1],
    translation_vector[2]
    )

# We are now ready to blend the walk and death animations
blend_animator = generate_blend_animator(walk_end_for_blend, death_start_for_blend, blend_curve)

# We finish by merging the 3 animations into a single one
# Walk start -> Walk end blended with death start -> Death end
clipped_animator = walk_start.clip(blend_animator).clip(death_end)

# We save the resulting animation to a file
serialize(clipped_animator, "walk_then_suddenly_die", SERIALIZED_DATA_DIR)

# It is now ready to be used in the GUI
# Result is shown in the video Features_deno.mp4


import unittest
import numpy as np
import quaternion

from skeletal_animation.core.model.joint import Joint
from skeletal_animation.core.math.transform import Transform, identity
from skeletal_animation.core.math.constants import HIGH_R_TOL, HIGH_A_TOL

class TestJointDefaultConstructor(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back", local_bind_transform=Transform(pos=[1,0,1]))
        self.root.add_child(self.back)
        self.arm = Joint(name="arm", local_bind_transform=Transform(pos=[0.25,0,0], orient=np.quaternion(1,0,0,0)))
        self.back.add_child(self.arm)
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            self.hand = Joint(name="hand", local_bind_transform=Transform(orient=quaternion.from_float_array([.2,.8,.2,0])))
        self.arm.add_child(self.hand)
        # root -> back -> arm -> hand
        
    def test_default_constructor_with_default_parameters(self):
        j = Joint()
        self.assertIsNone(j.parent)
        self.assertEqual(j.name, "new joint")
        self.assertEqual(j.local_bind_transform, identity)

    def test_default_constructor_with_no_parent(self):
        t = Transform([1,0,2], np.quaternion(0, 2/3, 1/3, 2/3))
        j = Joint(name="root", local_bind_transform=t)
        self.assertIsNone(j.parent)
        self.assertEqual(j.name, "root")
        self.assertEqual(j.local_bind_transform, t)
    
    def test_default_constructor_with_root_parent(self):          
        j = self.back
        self.assertEqual(j.parent, self.root)
    
    def test_default_constructor_with_non_root_parent(self):  
        j = self.hand
        self.assertEqual(j.parent.parent.parent, self.root)


class TestJointIsCyclic(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back", local_bind_transform=Transform(pos=[1,0,1]))
        self.root.add_child(self.back)
        # root -> back
    
    def test_is_cyclic_simple(self):
        self.assertFalse(self.root.is_cyclic())
        with self.assertWarns(UserWarning):
            self.back.add_child(self.root)
        # root -> back -> root  =>  CYCLE!
        self.assertTrue(self.root.is_cyclic())
        self.back.remove_child(self.root)
        self.assertFalse(self.root.is_cyclic())

    def test_is_cyclic_intermediate(self):
        r = Joint(name="r")
        c1 = Joint(name="c1")
        c2 = Joint(name="c2")
        c3 = Joint(name="c3")
        r.add_children([c1, c2])
        c1.add_child(c3)
        self.assertFalse(r.is_cyclic())
        with self.assertWarns(UserWarning):
            c3.add_child(c1)
        self.assertFalse(r.is_cyclic())
        self.assertTrue(c1.is_cyclic())
        c3.remove_child(c1)
        self.assertFalse(c1.is_cyclic())
        c3.add_child(r)
        self.assertFalse(c1.is_cyclic())

    def test_is_cyclic_complex(self):
        arm = Joint(name="arm", local_bind_transform=identity)
        self.back.add_child(arm)
        forearm = Joint(name="forearm", local_bind_transform=identity)
        arm.add_child(forearm)
        hand = Joint(name="hand", local_bind_transform=identity)
        forearm.add_child(hand)
        finger = Joint(name="finger", local_bind_transform=identity)
        hand.add_child(finger)
        fingertip = Joint(name="fingertip", local_bind_transform=identity)
        finger.add_child(fingertip)
        head = Joint(name="head", local_bind_transform=identity)
        self.back.add_child(head)
        nose = Joint(name="nose", local_bind_transform=identity)
        head.add_child(nose)
        # root -> back -> arm -> forearm -> hand -> finger -> fingertip
        #              -> head -> nose
        
        self.assertFalse(self.root.is_cyclic())
        self.assertFalse(arm.is_cyclic())
        with self.assertWarns(UserWarning):
            finger.add_child(self.back)
        # root -> 
        # back -> arm -> forearm -> hand -> finger -> fingertip
        #      -> head -> nose -> arm              -> back -> arm -> forearm -> hand -> finger -> fingertip
        #                                                  -> head -> nose -> arm 
        # => CYCLE!  
        self.assertFalse(self.root.is_cyclic())      
        self.assertTrue(arm.is_cyclic())


class TestJointAddChild(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.arm = Joint(name="arm")
        # wanted hierarchy
        # root -> back -> arm
    
    def test_parent_update(self):
        self.assertIsNone(self.back.parent)
        self.root.add_child(self.back)
        self.assertEqual(self.back.parent, self.root)
        self.assertEqual(self.root.children, [self.back])
    
    def test_cyclicity_warning(self):
        self.root.add_child(self.back)
        self.back.add_child(self.arm)
        with self.assertWarns(UserWarning):
            self.arm.add_child(self.root)

class TestJointAddChildre(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.leftarm = Joint(name="leftarm")
        self.rightarm = Joint(name="rightarm")
        # wanted hierarchy
        # root -> back -> leftarm
        #              -> rightarm

    def test_joint_add_children(self):
        self.root.add_children([self.back])
        self.back.add_children([self.rightarm, self.leftarm])
        self.assertIsNone(self.root.parent)
        self.assertEqual(self.back.parent, self.root)
        self.assertEqual(self.leftarm.parent, self.back)
        self.assertEqual(self.rightarm.parent, self.back)

class TestJointRemoveChild(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.arm = Joint(name="arm")

    def test_parent_update(self):
        self.root.add_child(self.back)
        self.back.add_child(self.arm)
        self.root.remove_child(self.back)
        self.assertIsNone(self.back.parent)
        self.assertEqual(self.root.children, [])

class TestJointRemoveAllChildren(unittest.TestCase):
    def setUp(self): 
        self.root = Joint(name="root")
        self.back = Joint(name="back")
        self.leftarm = Joint(name="leftarm")
        self.rightarm = Joint(name="rightarm")

    def test_parent_update(self):
        self.root.add_child(self.back)
        self.back.add_child(self.leftarm)
        self.back.add_child(self.rightarm)        
        self.root.remove_all_children()
        self.assertIsNone(self.back.parent)        
        self.back.remove_all_children()
        self.assertEqual(self.back.children, [])

class TestJointForwardsKinematics(unittest.TestCase):
    def setUp(self):
        
        # Init joints
        self.root = Joint(name="root", local_bind_transform=identity)

        self.back = Joint(name="back", local_bind_transform=Transform(pos=[0,3,0]))

        self.leftarm_q = np.quaternion(-1, 0, 1, 0)
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):            
            self.leftarm_t = Transform(pos=[2, 0, 0], orient=self.leftarm_q)
        self.leftarm_t.normalize_quaternion()
        self.leftarm = Joint(name="leftarm", local_bind_transform=self.leftarm_t)

        rightarm_t = Transform(pos=[-2, 0, 0])
        self.rightarm = Joint(name="rightarm", local_bind_transform=rightarm_t) 

        self.lefthand_t = Transform(pos=[1, 0, 0])        
        self.lefthand = Joint(name="lefthand", local_bind_transform=self.lefthand_t)

        # Build hierarchy
        self.root.add_child(self.back)
        self.back.add_children([self.leftarm, self.rightarm])
        self.leftarm.add_child(self.lefthand)

    def test_joint_fk_with_no_pose_dict_and_root_at_identity(self):        
        self.assertEqual(self.root.local_bind_transform, identity)
        self.assertEqual(self.root.model_transform, None)
        self.assertEqual(self.leftarm.local_bind_transform, self.leftarm_t)
        self.assertEqual(self.leftarm.model_transform, None)
        self.assertEqual(self.lefthand.local_bind_transform, self.lefthand_t)
        self.assertEqual(self.lefthand.model_transform, None)

        self.root.forward_kinematics(current_pose_dict=None)

        self.assertEqual(self.root.local_bind_transform, identity)
        self.assertEqual(self.root.model_transform, identity)

        self.assertEqual(self.leftarm.local_bind_transform, self.leftarm_t)

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            leftarm_t = Transform(pos=[2,3,0], orient=self.leftarm_q)
        leftarm_t.normalize_quaternion()
        self.assertEqual(self.leftarm.model_transform, leftarm_t)

        self.assertEqual(self.lefthand.local_bind_transform, self.lefthand_t)
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            lefthand_t = Transform(pos=[2,3,1], orient=self.leftarm_q)
        lefthand_t.normalize_quaternion()
        self.assertEqual(self.lefthand.model_transform, lefthand_t)

    def test_joint_fk_with_no_pose_dict_and_root_translated(self):    
        new_root_t = Transform(pos=[0, -1, 1])
        self.root.local_bind_transform = new_root_t

        self.root.forward_kinematics(current_pose_dict=None)

        self.assertEqual(self.root.model_transform, new_root_t)

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            leftarm_t = Transform(pos=[2,2,1], orient=self.leftarm_q)
        leftarm_t.normalize_quaternion()
        self.assertEqual(self.leftarm.model_transform, leftarm_t)

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            lefthand_t = Transform(pos=[2,2,2], orient=self.leftarm_q)
        lefthand_t.normalize_quaternion()
        self.assertEqual(self.lefthand.model_transform, lefthand_t)
    
    def test_joint_fk_with_no_pose_dict_and_root_rotated(self):       
        new_q = np.quaternion(1, 0, 1+np.sqrt(2), 0) # right handed 135° rotation around the y-axis

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            new_root_t = Transform(orient=new_q)
        new_root_t.normalize_quaternion()
        self.root.local_bind_transform = new_root_t

        self.root.forward_kinematics(current_pose_dict=None)

        self.assertEqual(self.root.model_transform, new_root_t)


        leftarm_q = np.quaternion(1+np.sqrt(2), 0, 1, 0)
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            leftarm_t = Transform(pos=[-np.sqrt(2), 3, -np.sqrt(2)], orient=leftarm_q)
        leftarm_t.normalize_quaternion()
        self.assertEqual(self.leftarm.model_transform, leftarm_t)

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            lefthand_t = Transform(pos=[-np.sqrt(2)/2, 3, -3*np.sqrt(2)/2], orient=leftarm_q)
        lefthand_t.normalize_quaternion()
        self.assertEqual(self.lefthand.model_transform, lefthand_t)

    def test_joint_fk_with_translation_only_pose_dict_and_static_root_at_identity(self):
        # place root at identity
        self.root.local_bind_transform = identity

        pose_dict = {}
        pose_dict[self.root.id] = self.root.local_bind_transform # root doesn't move
        pose_dict[self.back.id] = Transform(pos=[0,1,0]) # back is lifted up
        pose_dict[self.leftarm.id] = identity # left arm doesn't move
        pose_dict[self.rightarm.id] = Transform(pos=[-3,-1,0]) # right arm extends and moves down
        pose_dict[self.lefthand.id] = Transform(pos=[2,0,0]) # left hand extends

        self.root.forward_kinematics(current_pose_dict=pose_dict)

        self.assertEqual(self.root.model_transform, identity)
        self.assertEqual(self.back.model_transform, Transform(pos=[0,4,0]))        
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            leftarm_t = Transform(pos=[2,4,0], orient=self.leftarm_q)
        leftarm_t.normalize_quaternion()
        self.assertEqual(self.leftarm.model_transform, leftarm_t)
        self.assertEqual(self.rightarm.model_transform, Transform(pos=[-5,3,0]))
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            lefthand_t = Transform(pos=[2,4,3], orient=self.leftarm_q)
        lefthand_t.normalize_quaternion()
        self.assertEqual(self.lefthand.model_transform, lefthand_t)

    def test_joint_fk_with_pose_dict_and_static_root_rotated(self):    
        # rotate root
        self.root.local_bind_transform = Transform(orient=quaternion.y) # 180° rotation around the y-axis (quaternion.y == np.quaternion(0, 0, 1, 0))

        pose_dict = {}
        pose_dict[self.root.id] = self.root.local_bind_transform # root doesn't move
        pose_dict[self.back.id] = Transform(pos=[0,1,0]) # back is lifted up  
        pose_dict[self.leftarm.id] = Transform(orient=np.quaternion(np.sqrt(2)/2, 0, 0, np.sqrt(2)/2)) # left arm rotates up (left handed 90° rotation around the pre rotation x-axis which is now the z-axis)
        pose_dict[self.rightarm.id] = identity # right arm doesn't move
        pose_dict[self.lefthand.id] = Transform(pos=[-0.5,0,0]) # left hand shrinks

        self.root.forward_kinematics(current_pose_dict=pose_dict)

        self.assertEqual(self.root.model_transform, Transform(orient=quaternion.y))
        self.assertEqual(self.back.model_transform, Transform(pos=[0,4,0], orient=quaternion.y))        
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            leftarm_t = Transform(pos=[-2,4,0], orient=np.quaternion(1, 1, 1, 1))
        leftarm_t.normalize_quaternion()
        self.assertEqual(self.leftarm.model_transform, leftarm_t)
        self.assertEqual(self.rightarm.model_transform, Transform(pos=[2,4,0], orient=quaternion.y))        
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):       
            lefthand_t = Transform(pos=[-2,4.5,0], orient=np.quaternion(1, 1, 1, 1))
        lefthand_t.normalize_quaternion()
        self.assertEqual(self.lefthand.model_transform, lefthand_t)
    
    def test_joint_fk_with_pose_dict_and_animated_root(self):
        # place root at identity
        self.root.local_bind_transform = identity

        new_q = np.quaternion(1, 0, 1+np.sqrt(2), 0).normalized() # right handed 135° rotation around the y-axis

        pose_dict = {}        
        pose_dict[self.root.id] = Transform(pos=[0,0,1], orient=new_q) # root moves "towards us" and rotates along the y-axis
        pose_dict[self.back.id] = Transform(pos=[0,1,0]) # back is lifted up
        pose_dict[self.leftarm.id] = identity # left arm rotates up (left handed 90° rotation around the pre rotation x-axis which is now the z-axis)
        pose_dict[self.rightarm.id] = Transform(pos=[-3,0,0]) # right arm extends
        pose_dict[self.lefthand.id] = identity # left hand shrinks

        self.root.forward_kinematics(current_pose_dict=pose_dict)
        
        
        self.assertEqual(self.root.model_transform, Transform(pos=[0,0,1], orient=new_q))
        self.assertEqual(self.back.model_transform, Transform(pos=[0,4,1], orient=new_q))
        self.assertEqual(
            self.leftarm.model_transform,
            Transform(
                pos=[-np.sqrt(2), 4, -np.sqrt(2) + 1],
                orient=np.quaternion(1+np.sqrt(2), 0, 1, 0).normalized()
            )
        )
        self.assertEqual(self.rightarm.model_transform, Transform(pos=[5*np.sqrt(2)/2, 4, 5*np.sqrt(2)/2 + 1], orient=new_q))            
        self.assertEqual(
            self.lefthand.model_transform,
            Transform(
                pos=[-np.sqrt(2)/2, 4, -3*np.sqrt(2)/2 + 1],
                orient=np.quaternion(1+np.sqrt(2), 0, 1, 0).normalized()
            )
        )



    


    



import numpy as np

from skeletal_animation.core.math import Transform
from skeletal_animation.core.model import Joint, Skeleton
from skeletal_animation.core.animate import *
from skeletal_animation.utils.loader import serialize

# CREATING A SKELETON
# ------------------
# Create a root joint with name "root" and default parameters
root = Joint(name="root")

# Create a skeleton with the root joint as root
skeleton = Skeleton(root)

# Create a joint with name "back" and default parameters
back = Joint(name="back")
root.add_child(back) # Add the joint as a child of the root joint

# We only need set position of the back joint, not its orientation
back.local_bind_transform.pos = [0,50,0]

# CREATING MOTION
# ---------------
# Create a translation curve for the root joint
root_translation_curve = Curve()

# Create a key at time 0 with a position of (0,0,0)
key0 = Key(time=0., value=[0.,0.,0.]) # all values must be given as floats
# Create a key at time 1 with a position of (0,0,100)
key1 = Key(time=1., value=[0.,0.,100.])
# Add the keys to the translation curve
root_translation_curve.add_key(key0)
root_translation_curve.add_key(key1)

# Create a joint animation for the root joint and give it our translation curve
root_joint_animation = JointAnimation(root, translation_curve=root_translation_curve)

# Create an animation and give it our joint animation
animation = Animation(joint_animations=[root_joint_animation])

# Create an animator object and give it our animation and skeleton
animator = Animator(animation, skeleton, kps=1.)

# NEXT STEPS
# ----------
# The head joint is located 15 units above its parent
head = Joint(name="head", local_bind_transform=Transform(pos=[0.,15.,0.]))
# The right arm joint is located 30 units to the right of its parent
right_arm = Joint("right_arm", Transform(pos=[-30.,0.,0.]))
# The left arm joint is located 30 units to the left of its parent
left_arm = Joint("left_arm", Transform(pos=[30.,0.,0.]))
# The right forearm joint is located 20 units to the right of its parent
left_forearm = Joint("left_forearm", Transform(pos=[-20.,0.,0.]))
# The left forearm joint is located 20 units to the left of its parent
right_forearm = Joint("right_forearm", Transform(pos=[20.,0.,0.]))

# Add the head joint as a child of the back joint
back.add_child(head)
# Add the right arm joint as a child of the back joint
back.add_child(right_arm)
# Add the left arm joint as a child of the back joint
back.add_child(left_arm)
# Add the right forearm joint as a child of the right arm joint
right_arm.add_child(left_forearm)
# Add the left forearm joint as a child of the left arm joint
left_arm.add_child(right_forearm)

# Create a key at time 1 with no rotation
key0 = Key(1., np.quaternion(1, 0, 0, 0))
# Create a key at time 3 with a rotation of ~53 degrees around the z axis
key1 = Key(3., np.quaternion(0.5, 0, 0, -1).normalized())
# Do not forget to normalize the quaternions when necessary
# Create a key at time 4 with a rotation of ~127 degrees around the z axis
key2 = Key(4., np.quaternion(2, 0, 0, -1).normalized())
# Create a key at time 5 with a rotation of ~53 degrees around the z axis
key3 = Key(5., np.quaternion(0.5, 0, 0, -1).normalized())
# Create a key at time 7 with no rotation
key4 = Key(7., np.quaternion(1, 0, 0, 0))

# Create a rotation curve for the right arm joint
right_arm_rotation_curve = Curve(keys=[key0, key1, key2, key3, key4])

# First we get the maximum time of the curve
max_time = right_arm_rotation_curve.get_max_key_time()
# Then we normalize the curve by this maximum time
right_arm_rotation_curve.normalize_times(max_time)
# This effectively divides all the key times by the maximum time

# Create a joint animation for the right arm joint
right_arm_animation = JointAnimation(
    joint=right_arm,
    rotation_curve=right_arm_rotation_curve
    )
# Add the joint animation to the animation
animation.joint_animations.append(right_arm_animation)

# Remove the key at time 1 of the root's translation curve
root_joint_animation.translation_curve.remove_key_at(1.)

# Replace default path with the path where you want to save the serialized animation
serialize(
    animator,
    "demo_hello_world",
    "D:/He-Arc/TB/tb-animation-squelettale/skeletal_animation/animated_models/serialized/"
    )




# Low tolerance are used for comparison that require high precision.
# e.g. verifying that a transform is exactly equal to another transform.

LOW_R_TOL = 1e-10 # Low relative tolerance for high-precision floating point comparisons
LOW_A_TOL  = 1e-14 # Low absolute tolerance for high-precision floating point comparisons

# High tolerance are used for comparisons that do not require high precision.
# e.g. verifying that a test result for transform multiplication is close to
# its expected result given by an external tool.

HIGH_R_TOL = 1e-5 # High relative tolerance for low-precision floating point comparisons
HIGH_A_TOL = 1e-7 # High absolute tolerance for low-precision floating point comparisons

# Very high tolerance are used for comparisons that require low precision.
# e.g. verifying that a key exist at a given time before adding a new one.
# If a user wants to add a key at time 0.10000 and a key exist at time 0.10001,
# the user should be prevented from adding the key. The keys are so close that
# a human would not be able to tell the difference in the animation. Therefore,
# adding the key would be irrelevent.

VERY_HIGH_R_TOL = 1e-3 # Very high relative tolerance for very low-precision floating point comparisons
VERY_HIGH_A_TOL = 1e-4 # Very high absolute tolerance for very low-precision floating point comparisons


import numpy as np
import quaternion

from skeletal_animation.core.animate import Curve
from skeletal_animation.core.math import Transform

class JointAnimation:
    """Class to store a joint's animation curves.

    It contains the animation curves for the position and orientation of the joint.
    As well as the time warp curve.

    The time warp curve is a curve that can be used to control the speed of the animation
    as a function of time.
    It contains keys with a single float as value called timekeys.
    - The time of the timekey is the real time of the animation.
    - The value of the timekey is the time at which the translation and rotation curves are evaluated.

    let :math:`R(t)` the result of evaluating the rotation curve at time :math:`t` and
    :math:`f(t)` the result of evaluating the time warp curve at time :math:`t`.

    Then the effective rotation of the joint at time :math:`t` is:
    :math:`R(f(t))`
    in other words, the rotation of the joint at time :math:`t` is the result of evaluating the rotation curve at :math:`f(t)` 
    This works similarly for translations.

    The logic for time warping is applied in the get_local_transform method.

    Examples
    -------
    # A translation curve that goes from [0,0,0] to [1,1,1] at time 0 and 1 respectively.
    >>> translation_curve = Curve(
    >>>     [Key(time=0., value=[0,0,0]),
    >>>     Key(time=1., value=[1,1,1])]
    >>> )

    # A default time warp curve. The value of the key at time 0 is 0 and the value of the key at time 1 is 1.
    >>> timewarp_curve = Curve(
    >>>    [Key(time=0., value=0.),
    >>>    Key(time=1., value=1.)]
    >>> )
        
    # A JointAnimation object with the above curves.
    >>> joint_animation = JointAnimation(
    >>>     joint=Joint(),
    >>>     translation_curve=translation_curve,
    >>>     rotation_curve=Curve(),
    >>>     timewarp_curve=timewarp_curve
    >>> )
    # let us see the result of the get_local_transform method at different times using linear interpolation.
    >>> for t in [0., 0.25, 0.5, 0.75, 1.]: 
    >>>    print(joint_animation.get_local_transform(t, 'linear').pos)
    [0,0,0]
    [0.25,0.25,0.25]
    [0.5,0.5,0.5]
    [0.75,0.75,0.75]
    [1,1,1]

    # Now if we add a key to the time warp curve...
    # Adding a time key at time 0.5 with value 0.2.
    >>> joint_animation.timewarp_curve.set_key_at(time=0.5, value=0.2)
    >>> # or joint_animation.timewarp_curve.add_key(Key(time=0.5, value=0.2))
    # This means that at time 0.5 of the animation, the translation and rotation curves are evaluated at time 0.2.
    >>> for t in [0., 0.25, 0.5, 0.75, 1.]: 
    >>>    print(joint_animation.get_local_transform(t, 'linear').pos)
    [0,0,0]
    [0.1,0.1,0.1]
    [0.2,0.2,0.2]
    [0.6,0.6,0.6]
    [1,1,1]
    # As we can see, the addition of this timekey causes the animation to start slow and speed up from t = 0.5.
    # In other words, 20% of the animation is performed in the first half of the animation and
    # the other 80% of the animation is performed in the second half of the animation.

    Attributes:
    ----------
    joint : Joint
        The joint that is animated.
    translation_curve : AnimationCurve
        The animation curve on the position of the joint.
    rotation_curve : AnimationCurve
        The animation curve on the orientation of the joint.
    timewarp_curve : AnimationCurve
        The time warp curve to apply to the animation.        
    """
    def __init__(self, joint, translation_curve=Curve(), rotation_curve=Curve(), timewarp_curve=Curve()):
        """**Default constructor for the JointAnimation class.**

        Parameters:
        ----------
        joint : Joint
            The joint that is animated.
        translation_curve : Curve, optional
            The animation curve on the position of the joint. By default, an empty curve is used.         
        rotation_curve : Curve, optional
            The animation curve on the orientation of the joint. By default, an empty curve is used.
        timewarp_curve : Curve, optional
            The time warp curve to apply to the animation. By default, a curve with a start key
            at time 0 with value 0 and an end key at time 1 with value 1 is used.

        Notes
        ----
        A time warp curve must contain a start and end key. The start key must have time 0 and the end key must have time 1.
        If no start key is found, a start key is added with time 0 and value 0.
        Similarly, if no end key is found, a end key is added with time 1 and value 1.
        This is done by the init_timewarp_curve method.
        """
        self.joint = joint
        self.translation_curve = translation_curve
        self.rotation_curve = rotation_curve
        self.timewarp_curve = timewarp_curve

        self.init_timewarp_curve()
    
    def init_timewarp_curve(self):
        """Initializes the timewarp curve with a start key at time 0 with value 0
        and an end key at time 1 with value 1.
        """
        self.timewarp_curve.set_key_at(time=0., value=0.)
        self.timewarp_curve.set_key_at(time=1., value=1.)


    def get_local_transform(self, time, interpolation_algorithm):
        """Computes the local animation transform for joint at the given time.

        Parameters:
        ----------
        time : float
            The time at which to compute the transform.
        interpolation_algorithm : str
            The algorithm to use for interpolating the animation curves.
            Possible values: 'constant' and 'linear'.

        Raises:
        ----------
        ValueError
            If the interpolation algorithm is not valid. 
        
        Returns:
        ----------
        Transform
            The local animation transform for the joint at the given time.

        Notes
        -----
        If the translation curve is empty, the behavior of this method changes whether
        the joint is a root joint or not. If the joint is a root joint, the translation
        is set to its local bind transform's position and if the joint is not a root joint,
        the translation is set to [0,0,0].
        Similarly, if the rotation curve is empty, the behavior of this method changes whether
        the joint is a root joint or not. If the joint is a root joint, the rotation is set to its
        local bind transform's orientation and if the joint is not a root joint, the rotation is set
        to the identity quaternion (1,0,0,0).
        If the timewarp curve is empty, this method behaves the same way as with a default timewarp curve.
        """
        if self.timewarp_curve.is_empty():
            key_time = time
        else:
            key_time = self.timewarp_curve.get_lerp_value(time)
        
        if interpolation_algorithm == "constant":
            translation = self.translation_curve.get_previous_value(key_time)
            rotation = self.rotation_curve.get_previous_value(key_time)
        elif interpolation_algorithm == "linear":
            translation = self.translation_curve.get_lerp_value(key_time)
            rotation = self.rotation_curve.get_lerp_value(key_time)
        elif interpolation_algorithm == "spline":
            raise NotImplementedError("Spline interpolation not implemented yet.")
        else:
            raise ValueError("Invalid interpolation algorithm. Must be one of 'constant', 'linear', or 'spline'.")

        if translation is None:
            if self.joint.parent is None:
                translation = self.joint.local_bind_transform.pos
            else:
                translation = np.array([0,0,0])
        
        if rotation is None:
            if self.joint.parent is None:
                rotation = self.joint.local_bind_transform.orient
            else:
                rotation = quaternion.one

        return Transform(pos=translation, orient=rotation)

    def __repr__(self):
        return f"JointAnimation(joint={str(self.joint)},\
        translation_curve={str(self.translation_curve)}, \
        rotation_curve={str(self.rotation_curve)},\
        timewarp_curve={str(self.timewarp_curve)})" 

    def __eq__(self, other):
        """Equality operator for JointAnimation objects.

        Note:
        ----------
        Two JointAnimation objects are considered equal if:
        - Their joints are equal. (See Joint.__eq__)
        - Their translation curves are equal. (See Curve.__eq__)
        - Their rotation curves are equal. (See Curve.__eq__)
        - Their timewarp curves are equal. (See Curve.__eq__)

        Parameters
        ----------
        other : JointAnimation
            JointAnimation object to compare to.

        Returns
        -------
        bool
            True if the two JointAnimation objects are equal, False otherwise.
        """
        return (self.joint == other.joint
            and self.translation_curve == other.translation_curve
            and self.rotation_curve == other.rotation_curve
            and self.timewarp_curve == other.timewarp_curve)


import json
import copy
import warnings

from skeletal_animation.core.math import Transform, identity

class Joint:
    """A class to represent a joint.

    A joint is considered a root if it has no parent.
    
    Attributes
    ----------
    id : int
        Unique id of the joint.
    name : str
        Name given to the joint. Not necessarily unique.    
    parent : Joint
        Parent of the joint.
    children : (Joint) list
        List of children joints.
    local_bind_transform : Transform
        Initial transform of the joint in the parent's local coordinate system.
    model_transform : Transform
        Current transform of the joint in the model (or world) coordinate system.

    Static Attributes
    ------------------
    LAST_ID : int
        Auto incrementing ID for the next joint.
    """
    LAST_ID = 0

    
    def __init__(self, name="new joint", local_bind_transform=identity):
        """**Default constructor for the Joint class.**

        Parameters
        ----------
        name : str, optional
            Name of the joint, by default "new joint".
        local_bind_transform : Transform, optional
            Initial transform relative to the parent, by default identity transform.
        """
        self.id = Joint.LAST_ID
        Joint.LAST_ID += 1
        self.name = name
        self.parent = None   
        self.children = []
        self.local_bind_transform = local_bind_transform
        self.model_transform = None 
    
    
    def add_child(self, joint):
        """Makes the given joint a child of the current joint.

        Note
        ----
        The joint's parent is updated automatically.

        Parameters
        ----------
        joint : Joint
            The joint to be added as a child.

        Raises
        --------
        If the addition of the joint makes the tree cyclic, an error is raised.        
        """
        if joint.parent is None:
            joint.parent = self
            self.children.append(joint)
        elif joint.parent != self:
            old_parent = joint.parent
            old_parent.children.remove(joint)
            joint.parent = self
            self.children.append(joint)
        if self.is_cyclic():              
            warnings.warn("The addition of the joint makes the tree cyclic.", UserWarning)

       
    def add_children(self, joints):
        """Makes the given joints children of the current joint.

        Note
        ----
        The joints' parents are updated automatically.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be added as children.
        """
        for joint in joints:
            self.add_child(joint)

    
    def remove_child(self, joint):
        """Removes the given joint from the children of the current joint.

        Note
        ----
        If the joint is not a child of the current joint, then nothing happens.
        The joint's parent is updated automatically.
        
        Parameters
        ----------
        joint : Joint
            The joint to be removed.
        """
        if joint in self.children:
            self.children.remove(joint)
            joint.parent = None

    
    def remove_children(self, joints):
        """Removes the given joints from the children of the current joint.

        Note
        ----
        The joints' parents are updated automatically.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be removed.
        """
        for joint in joints:
            self.remove_child(joint)

    
    def remove_all_children(self):
        """Removes all the children of the current joint.

        Note
        ----
        The removed joints' parents are updated automatically.
        """
        self.remove_children(copy.copy(self.children))


    def forward_kinematics(self, current_pose_dict, parent_model_transform=identity):
        """Computes the model transform of the current joint and all its children recursively using forward kinematics.

        The formula used to compute the model transform of the current joint is:
        model_transform = parent_model_transform x local_bind_transform x local_anim_transform.

        This formula is valid for transform matrices in column-major order, as is the case in this module.

        Parameters
        ----------
        current_pose_dict : Dict[int, Transform]
            Dictionary with joint id as keys and their respective transform as values.
            Dictionary containing the local transforms to be applied to each joint in the skeleton for it 
            to be in the current pose, also known as the local animation transforms.
            It is obtained from the animation data stored in curves. 
            See generate_pose() method in the Animator class.
        parent_model_transform : Transform, optional
            Model transform of the parent joint, by default identity transform. 
            This parameter is optional only because the parent model transform for the root joint 
            does not exist. In that case, it is set to identity.
        """
        # If current_pose_dict is None (i.e. when compute_bind_pose() is called),
        # then the root's transform is its local bind transform
        # and the delta for other joints is null (identity transform).

        current_local_transform = Transform()

        if current_pose_dict == None:
            if self.parent is None: # root
                current_local_transform = self.local_bind_transform
            else:
                # current_local_transform = Transform()
                pass
        else:
            if self.id in current_pose_dict:
                current_local_transform = current_pose_dict[self.id]   

        # Assumes that the actual transform is given for the root joint
        # And that, for other joints, the delta transform is given instead.

        if self.parent is None: # root
            current_model_transform = current_local_transform
        else:
            current_model_transform = parent_model_transform @ self.local_bind_transform @ current_local_transform
        
        self.model_transform = current_model_transform  

        # Recursively compute the model transforms of the children
        for child_joint in self.children:
            child_joint.forward_kinematics(current_pose_dict, current_model_transform)        

    
    def is_cyclic(self):
        """Checks for cyclicity in the joint hierarchy for the current joint.

        Note: The joint hierarchy is considered cyclic only if a joint is the parent of one of its ancestors.

        Examples
        -------
        >>> root = Joint()
        >>> child1 = Joint()
        >>> child2 = Joint()
        >>> child3 = Joint()
        >>> root.add_children([child1, child2])
        >>> child1.add_child(child3)
        # The joint hierarchy is now:
        #   root -> child1 -> child3
        #        -> child2
        >>> root.is_cyclic()
        False
        >>> child3.add_child(child1)
        # The joint hierarchies are now:
        # root -> child2  and  ... -> child1 -> child3 -> child1 -> child3 -> child1 -> ...
        >>> root.is_cyclic()
        False
        >>> child1.is_cyclic()
        True
        
        Returns
        -------
        bool
            True if the joint hierarchy is cyclic, False otherwise.
        """

        # List of visited nodes.
        visited = []
        # Stack to keep track of the current node.
        stack = []
        stack.append(self)

        # Run till stack is empty or cycle is found.
        while stack:
            node = stack.pop()
            for child in node.children:
                if child in visited:
                    return True
                else:
                    stack.append(child)
                    visited.append(node)

        # If we reach here, then there is no cycle
        return False

    
    def __repr__(self):
        dict = {}
        dict["id"] = self.id
        dict["name"] = self.name
        if self.parent is None:
            dict["parent"] = None
        else:
            dict["parent"] = self.parent.name
        dict["children"] = [child.name for child in self.children]
        dict["local_bind_transform"] = self.local_bind_transform.__repr__()
        if self.model_transform is None:
            dict["model_transform"] = None
        else:
            dict["model_transform"] = self.model_transform.__repr__()
        return json.dumps(dict, indent=4)

    def __eq__(self, other):
        """Equality operator for joint objects.

        Note
        ----
        Two joints are considered equal if:
        - They are of the same type (i.e. have the same class)
        - They have the same id, name, parent and children
        - Their local bind transforms are equal. (See Transform.__eq__())

        Model transforms are not necessarily equal, since the model transform
        of a same joint changes over the course of the animation.

        Parameters
        ----------
        other : Joint
            Joint object to compare to.

        Returns
        -------
        bool
            True if the two joints are equal, False otherwise.
        """
        if type(self) != type(other):
            return False

        if self.id != other.id or self.name != other.name or self.local_bind_transform != other.local_bind_transform:
            return False

        self_p_id = None
        other_p_id = None
        if self.parent is not None:
            self_p_id = self.parent.id
        if other.parent is not None:
            other_p_id = other.parent.id
        if self_p_id != other_p_id:
            return False

        if self.children != other.children:
            return False

        return True



import gh_utils
import gh_camera
import gh_renderer
import gh_gpu_program
import gh_object
import gh_imgui

from inspect import getmembers, isfunction

from skeletal_animation.core.math import *
from skeletal_animation.gui.init import SERIALIZED_ANIMATION_PATH
from skeletal_animation.utils.loader import *
from skeletal_animation.algo.motionblending import *

elapsed_time = gh_utils.get_elapsed_time()
dt = elapsed_time - last_time
last_time = elapsed_time

if (elapsed_time - g_fps_last_time) > 1.0:
    g_fps_last_time = elapsed_time
    g_fps = g_frames
    g_frames = 0

g_frames = g_frames + 1

# Background
#
gh_renderer.set_depth_test_state(0)

gh_camera.bind(camera_ortho)

gh_renderer.clear_color_depth_buffers(0, 0, 0, 0, 1.0)

gh_gpu_program.bind(color_prog)

gh_mesh.set_vertex_color(fullscreen_quad, 0, bkg_color_bottom.r, bkg_color_bottom.g, bkg_color_bottom.b, bkg_color_bottom.a) # bottom-left
gh_mesh.set_vertex_color(fullscreen_quad, 1, bkg_color_top.r, bkg_color_top.g, bkg_color_top.b, bkg_color_top.a) # top-left
gh_mesh.set_vertex_color(fullscreen_quad, 2, bkg_color_top.r, bkg_color_top.g, bkg_color_top.b, bkg_color_top.a) # top-right
gh_mesh.set_vertex_color(fullscreen_quad, 3, bkg_color_bottom.r, bkg_color_bottom.g, bkg_color_bottom.b, bkg_color_bottom.a) # bottom-right

gh_object.render(fullscreen_quad)


# Main rendering
#
gh_renderer.set_depth_test_state(1)

# Apply 3D camera settings
#
if not imgui_window_hovered:
  gh_camera.set_fov(camera, camera_fov)
  gx_camera.set_keyboard_speed(keyboard_speed)
  gx_camera.update(camera, dt)

gh_utils.tripod_visualizer_camera_render(camera, 0, 0, 100, 100)
gh_camera.bind(camera)

# The lighting shader
#
prog = lighting_prog
gh_gpu_program.bind(prog)
gh_gpu_program.uniform1i(prog, "tex0", 0)

cx, cy, cz = gh_camera.get_position(camera)

light0_pos = {"x":-160.0, "y":80.0, "z":-160.0}
light1_pos = {"x":160.0, "y":160.0, "z":80.0}

gh_gpu_program.uniform4f(prog, "light_ambient", 0.2, 0.2, 0.2, 1.0)
gh_gpu_program.uniform4f(prog, "light_specular", 0.6, 0.6, 0.6, 1.0)

gh_gpu_program.uniform4f(prog, "light0_position", light0_pos["x"], light0_pos["y"], light0_pos["z"], 1.0)
gh_gpu_program.uniform4f(prog, "light0_diffuse", 1.0, 1.0, 1.0, 1.0)

gh_gpu_program.uniform4f(prog, "light1_position", light1_pos["x"], light1_pos["y"], light1_pos["z"], 1.0)
gh_gpu_program.uniform4f(prog, "light1_diffuse", 1.0, 1.0, 1.0, 1.0)

gh_gpu_program.uniform4f(prog, "uv_tiling", 1.0, 1.0, 0.0, 1.0)

gh_gpu_program.uniform4f(prog, "material_diffuse", 1.0, 1.0, 1.0, 1.0)
gh_gpu_program.uniform4f(prog, "material_ambient", 1.0, 1.0, 1.0, 1.0)
gh_gpu_program.uniform4f(prog, "material_specular", 0.1, 0.1, 0.1, 1.0)
gh_gpu_program.uniform1f(prog, "material_shininess", 24.0)

gh_renderer.solid()

gh_object.render(grid)

LEFT_BUTTON = 1
mouse_left_button = gh_input.mouse_get_button_state(LEFT_BUTTON)
RIGHT_BUTTON = 2
mouse_right_button = gh_input.mouse_get_button_state(RIGHT_BUTTON)
mouse_x, mouse_y = gh_input.mouse_get_position()
IMGUI_WIDGET_SEPARATOR = 1
IMGUI_WIDGET_SAME_LINE = 2
IMGUI_WIDGET_BULLET = 3
IMGUI_WIDGET_VERTICAL_SPACING = 4

##_______________________________ANIMATION__________________________________##
# TODO better way than using global?
def load_animator_1():
    global animation_render_dict, SERIALIZED_ANIMATION_PATH, animated_model, anim_1, file_not_found_message
    #gh_utils.trace("Loading animator 1")
    try:
        file_not_found_message = None
        animator = load_serialized(animated_models[anim_1], SERIALIZED_ANIMATION_PATH)
        joint_list = animator.skeleton.as_joint_list()
        animator.skeleton.compute_bind_pose()
        animation_render_dict['anim1'] = [JointWrapper(joint) for joint in joint_list]
        return animator

    except FileNotFoundError:
        file_not_found_message = "File not found: " + SERIALIZED_ANIMATION_PATH + animated_models[anim_1] + ".pkl. Element removed from list."
        animated_models.pop(anim_1)
        anim_1 = 0
        return None

def load_animator_2():    
    global animation_render_dict, SERIALIZED_ANIMATION_PATH, animated_models, anim_2, file_not_found_message
    #gh_utils.trace("Loading animator 2")
    try:
        file_not_found_message = None
        animator = load_serialized(animated_models[anim_2], SERIALIZED_ANIMATION_PATH)
        joint_list = animator.skeleton.as_joint_list()
        animator.skeleton.compute_bind_pose()
        animation_render_dict['anim2'] = [JointWrapper(joint) for joint in joint_list]
        return animator

    except FileNotFoundError:
        file_not_found_message = "File not found: " + SERIALIZED_ANIMATION_PATH + animated_models[anim_2] + ".pkl. Element removed from list."
        animated_models.pop(anim_2)
        anim_2 = 0
        return None

def load_blend_animator():
    #gh_utils.trace("Loading blend animator")
    global animation_render_dict, weight, blend_error_message
    ba = generate_blend_animator(animator_1, animator_2, Curve([Key(0.0, weight)]))
    if ba is None:
        blend_error_message = "Cannot generate blend animation: incompatible skeletons"
    else:
        blend_error_message = None

        animator_2.duration = animator_1.duration
        ba.duration = animator_1.duration
        animator_2.anim_time = animator_1.anim_time
        ba.anim_time = animator_1.anim_time   

        joint_list = ba.skeleton.as_joint_list()
        ba.skeleton.compute_bind_pose()    
        animation_render_dict['blend'] = [JointWrapper(joint, Color(90, 180, 90)) for joint in joint_list]

    return ba

def update_animation(dt, stepping=False):
    # start = time.time()
    if is_playing or stepping:

        if animator_1 is not None:        
            animator_1.kps = anim_speed * 30
            if render_anim_1:
                animator_1.step(dt)
            else:
                animator_1.update(dt)
        
        if animator_2 is not None:
            animator_2.kps = anim_speed * 30
            if render_anim_2:
                animator_2.step(dt)
            else:
                animator_2.update(dt)

        if blend_animator is not None:
            blend_animator.kps = anim_speed * 30
            if render_blend:
                blend_animator.step(dt)
            else:
                blend_animator.update(dt)
    # end = time.time()
    # gh_utils.trace("update_animation : " + str(end - start))

def render_animation():
    global animation_render_dict, render_anim_1, render_anim_2, render_blend
    if render_anim_1:
        for joint_wrapper in animation_render_dict['anim1']:
            joint_wrapper.apply_transform()
            gh_object.render(joint_wrapper.get_id())
    if render_anim_2:
        for joint_wrapper in animation_render_dict['anim2']:
            joint_wrapper.apply_transform()
            gh_object.render(joint_wrapper.get_id())
    if render_blend:
        for joint_wrapper in animation_render_dict['blend']:
            joint_wrapper.apply_transform()
            gh_object.render(joint_wrapper.get_id())

def any_active_animation():
    global animator_1, animator_2, blend_animator
    return animator_1 is not None or animator_2 is not None or blend_animator is not None

def play():
    global is_playing
    is_playing = True

def pause():
    global is_playing
    is_playing = False


##_________________________________GUI______________________________________##
##
gh_imgui.frame_begin(winW, winH, mouse_x, mouse_y,
                     mouse_left_button, mouse_right_button, dt)


# Flags for window style, window position and window size.
##
window_default = 0
window_no_resize = 2
window_no_move = 4
window_no_collapse = 32
window_show_border = 128
window_no_save_settings = 256
pos_size_flag_always = 1  # Always set the pos and/or size
# Set the pos and/or size once per runtime session (only the first call with succeed)
pos_size_flag_once = 2
# Set the pos and/or size if the window has no saved data (if doesn't exist in the .ini file)
pos_size_flag_first_use_ever = 4
# Set the pos and/or size if the window is appearing after being hidden/inactive (or the first time)
pos_size_flag_appearing = 8


# Beginning of the window with caption "Menu"

window_flags = 0  # default flag

IMGUI_WINDOW_BG_COLOR = 1
gh_imgui.set_color(IMGUI_WINDOW_BG_COLOR, 0.1, 0.1, 0.1, 0.8)

is_open = gh_imgui.window_begin("Menu", 300, 200, 50, 20, window_flags,
                                pos_size_flag_first_use_ever, pos_size_flag_first_use_ever)

# The following test is important because without it, if we move the window, we will move the camera too...
##
if gh_imgui.is_window_hovered() or gh_imgui.is_any_item_hovered():
    imgui_window_hovered = True
else:
    imgui_window_hovered = False

if is_open == 1:
    window_w = gh_imgui.get_content_region_available_width()
    widget_width = window_w * 1.0

    gh_imgui.text("Press [ESC] to quit the demo")

    raw_fps_str = "Framerate: {fps:.0f} FPS"
    fps_str = raw_fps_str.format(fps = g_fps)
    gh_imgui.text(fps_str)


    import_button = gh_imgui.button("Import...", 200, 20)

    if import_button:
        gh_imgui.file_browser_open()
        is_file_browser_open = True

    if is_file_browser_open:
        gh_imgui.file_browser_display(600, 400)

        if gh_imgui.file_browser_has_selected():
            filename = gh_imgui.file_browser_get_selected()
            temp_animator = load_fbx(filename)
            anim_name = filename.split("\\")[-1].split(".")[0]
            serialize(temp_animator, anim_name, SERIALIZED_ANIMATION_PATH)
            animated_models.append(anim_name)
            gh_imgui.file_browser_close()
            is_file_browser_open = False

    anim_1_selector = gh_imgui.combo_box_create("Animation 1")
    for animated_model in animated_models:
        gh_imgui.combo_box_add_item(anim_1_selector, animated_model)

    anim_1 = gh_imgui.combo_box_draw(
        anim_1_selector,
        anim_1
    )

    anim_2_selector = gh_imgui.combo_box_create("Animation 2")
    for animated_model in animated_models:
        gh_imgui.combo_box_add_item(anim_2_selector, animated_model)

    anim_2 = gh_imgui.combo_box_draw(
        anim_2_selector,
        anim_2
    )

    anim_speed = gh_imgui.slider_1f(
        "Animation speed",
        anim_speed,
        0.1, 10.0,
        2.0
    )

    if anim_1 != previous_anim_1:
        if anim_1 != 0:
            animator_1 = load_animator_1()
            if anim_2 != 0:
                blend_animator = load_blend_animator()
        else:
            animator_1 = None

    if anim_2 != previous_anim_2:
        if anim_2 != 0:
            animator_2 = load_animator_2()
            if anim_1 != 0:
                blend_animator = load_blend_animator()
        else:
            animator_2 = None

    gh_imgui.begin_disabled(
        not any_active_animation()
    )

    play_pause_label = "Pause" if is_playing else "Play"
    play_pause_button = gh_imgui.button(play_pause_label, 200, 20)

    if play_pause_button:
        is_playing = not is_playing

    gh_imgui.begin_disabled(
        is_playing
    )

    stepping_rate = 30 # frames per second

    step_back_button = gh_imgui.button("Step back", 200, 20)
    if step_back_button:
        update_animation(-dt * g_fps / stepping_rate, stepping = True)
        render_animation()

    step_forward_button = gh_imgui.button("Step forward", 200, 20)
    if step_forward_button:
        update_animation(dt * g_fps / stepping_rate, stepping = True)
        render_animation()

    gh_imgui.end_disabled()

    if blend_error_message is not None:
        gh_imgui.text_rgba("(!) " + str(blend_error_message), 1.0, 0.0, 0.0, 1.0)
    if file_not_found_message is not None:
        gh_imgui.text_rgba("(!) " + str(file_not_found_message), 1.0, 0.0, 0.0, 1.0)

    gh_imgui.begin_disabled(anim_1 == 0 or anim_2 == 0)

    render_anim_1 = gh_imgui.checkbox("Anim 1", render_anim_1)
    render_anim_2 = gh_imgui.checkbox("Anim 2", render_anim_2)

    gh_imgui.begin_disabled(blend_error_message != None)

    render_blend = gh_imgui.checkbox("Blend", render_blend)

    weight = gh_imgui.slider_1f(
        "blending weight",
        weight,
        0.0, 1.0,
        1.0
    )

    gh_imgui.end_disabled()

    gh_imgui.end_disabled()

    gh_imgui.end_disabled()

if anim_1 != 0 and anim_2 == 0:
    render_anim_1 = True
    render_anim_2 = False
    render_blend = False
elif anim_1 == 0 and anim_2 != 0:
    render_anim_2 = True
    render_anim_1 = False
    render_blend = False
elif anim_1 != 0 and anim_2 != 0:
    if len(set(previous_weights)) > 1:
        blend_animator = load_blend_animator()
        blend_animator.step(0)
        previous_weights = [weight]

if is_playing:
    update_animation(dt)

render_animation()

# End of the window.
##
gh_imgui.window_end()

gh_imgui.frame_end()

previous_anim_1 = anim_1
previous_anim_2 = anim_2

previous_weights.append(weight)
if len(previous_weights) > 10:
    previous_weights.pop(0)

##_______________________________LIVE_CODING_______________________________##
# The live coding section must be empty when the program is loaded
# For things that will be executed every frame, just write them here.
# For things that will be executed only once, wrap them in a function.












































##_____________________________END_LIVE_CODING_____________________________##

if not function_list_loaded:
    previous_function_list = getmembers(sys.modules[__name__], isfunction)
    function_list_loaded = True

function_list = getmembers(sys.modules[__name__], isfunction)

function_buffer = [function for function in function_list if function[0] not in \
    [previous_function[0] for previous_function in previous_function_list]]

previous_function_list = function_list

for function in function_buffer:
    gh_utils.trace("Calling function: " + str(function))
    function[1]()
    del function





import numpy as np
import quaternion
import warnings

from skeletal_animation.core.math.constants import LOW_A_TOL, LOW_R_TOL

class Transform:
    """A class to represent a Transformation in 3D space composed by
    a translation and a rotation.

    np.quaternion is used to represent quaternions.

    Here are a few static values to declare common quaternions:

    - quaternion.one = np.quaternion(1, 0, 0, 0) -> identity quaternion

    - quaternion.x = np.quaternion(0, 1, 0, 0) -> 180 degree rotation around x axis

    - quaternion.y = np.quaternion(0, 0, 1, 0) -> 180 degree rotation around y axis

    - quaternion.z = np.quaternion(0, 0, 0, 1) -> 180 degree rotation around z axis
    
    Attributes
    ----------
    pos : ndarray (float)
        1D array containing the 3 values for the position in this order [x, y, z].
    orient : numpy quaternion
        Quaternion object defining the orientation.

        More info on numpy quaternion here: https://quaternion.readthedocs.io/en/latest/#
    identity : Transform
        Static attribute for the identity transform.

    Notes
    ----------
    Scale is not supported.
    """
    def __init__(self, pos = np.zeros(3), orient = quaternion.one):        
        """**Default constructor for the Transform class.**

        Parameters
        ----------
        pos : (float) ndarray, optional
            1D array containing the 3 values for the transform's position, by default [0, 0, 0].
        orient : np.quaternion, optional
            Quaternion object defining the transform's orientation, by default identity quaternion.
        """
        self.pos = pos
        self.orient = orient
        if not self.quaternion_is_unit():
            warnings.warn("Transform built with non unit quaternion. A normalization might be necessary. See Transform.normalize_quaternion().", stacklevel=2)
    
    @classmethod
    
    def from_matrix(cls, matrix=np.eye(4)):
        """Transform constructor from 4x4 transform matrix.

        Parameters
        ----------
        matrix : (float) ndarray, optional
            4x4 transformation matrix from which to extract position vector and orientation quaternion, by default 4x4 identity matri).

        Returns
        -------
        Transform
            Transform object.
        """                     
        matrix = np.array(matrix)
        if matrix.shape != (4,4) and matrix.shape != (3,3):            
            raise(IndexError("Improper array shape for extracing Quaternion. Expected array of shape (4,4) or (3,3) instead of {}".format(matrix.shape)))

        diagonal = matrix[0][0] + matrix[1, 1] + matrix[2, 2]
        if (diagonal > 0):
            w4 = np.sqrt(diagonal + 1) * 2
            w = w4 / 4
            x = (matrix[2, 1] - matrix[1, 2]) / w4
            y = (matrix[0, 2] - matrix[2, 0]) / w4
            z = (matrix[1, 0] - matrix[0, 1]) / w4
        elif ((matrix[0, 0] > matrix[1, 1]) and (matrix[0, 0] > matrix[2, 2])):
            x4 = np.sqrt(1 + matrix[0, 0] - matrix[1, 1] - matrix[2, 2]) * 2
            w = (matrix[2, 1] - matrix[1, 2]) / x4
            x = x4 / 4
            y = (matrix[0, 1] + matrix[1, 0]) / x4
            z = (matrix[0, 2] + matrix[2, 0]) / x4
        elif (matrix[1, 1] > matrix[2, 2]):
            y4 = np.sqrt(1 + matrix[1, 1] - matrix[0, 0] - matrix[2, 2]) * 2
            w = (matrix[0, 2] - matrix[2, 0]) / y4
            x = (matrix[0, 1] + matrix[1, 0]) / y4
            y = y4 / 4
            z = (matrix[1, 2] + matrix[2, 1]) / y4
        else:
            z4 = np.sqrt(1 + matrix[2, 2] - matrix[0, 0] - matrix[1, 1]) * 2
            w = (matrix[1, 0] - matrix[0, 1]) / z4
            x = (matrix[0, 2] + matrix[2, 0]) / z4
            y = (matrix[1, 2] + matrix[2, 1]) / z4
            z = z4 / 4

        v = matrix[0:3, 3]
        q = np.quaternion(w, x, y, z)
        return cls(v, q)

    
    def as_matrix(self):
        """Returns the 4x4 transformation matrix for the transform.
        
        Returns
        -------
        ndarray (float)
            4x4 transformation matrix.

        Note
        ----
        The matrix is in column-major order.
        """
        t_matrix = np.eye(4)
        t_matrix[0:3, 0:3] = quaternion.as_rotation_matrix(self.orient)
        
        t_matrix[0:3, 3] = self.pos
        return t_matrix

    
    def normalize_quaternion(self):
        """Normalizes the quaternion of the transform.
        """
        self.orient = self.orient.normalized()
    
    
    def quaternion_is_unit(self):
        """Checks if the quaternion for the transform's orientation is unitary.

        Returns
        -------
        bool
            True if the quaternion is unitary, False otherwise.
        """
        return np.isclose(np.linalg.norm(quaternion.as_float_array(self.orient)), 1, rtol=LOW_R_TOL, atol=LOW_A_TOL)
    
    
    def inversed(self):
        """Returns the inverse of the transform.

        Returns
        -------
        Transform
            Inverse transform.
        """
        return Transform.from_matrix(np.linalg.inv(self.as_matrix()))

    
    def multiply_by(self, other):
        """Generates a Transform that is the result of the multiplication of two transforms.

        Parameters
        ----------
        other : Transform
            Transform to multiply by.

        Returns
        -------
        Transform
            Resulting Transform. Corresponds to a transformation by the second matrix (other) followed by a transformation by the first matrix (self).
        """
        self_orient_norm = np.linalg.norm(quaternion.as_float_array(self.orient))
        other_orient_norm = np.linalg.norm(quaternion.as_float_array(other.orient))
        if not np.allclose([self_orient_norm, other_orient_norm], [1., 1.], rtol=LOW_R_TOL, atol=LOW_A_TOL):
            warnings.warn("Transform multiplication attempted with non unit quaternions. This can lead to erroneous results.", stacklevel=2)

        return Transform.from_matrix(np.matmul(self.as_matrix(), other.as_matrix())) 
    
    
    def translate(self, pos_vector):
        """Translates the transform by the given position vector.

        Parameters
        ----------
        pos_vector : ndarray (float)
            1D array containing the 3 values for the translation.

        Returns
        -------
        Transform
            Translated transform.
        """
        translation_transform = Transform(pos=pos_vector)
        return self.multiply_by(translation_transform)

    
    def rotate_quaternion(self, quaternion):
        """Rotates the transform by the given quaternion.

        Parameters
        ----------
        quaternion : np.quaternion
            Quaternion to rotate by.

        Returns
        -------
        Transform
            Rotated transform.
        """
        rotation_transform = Transform(orient=quaternion)
        return self.multiply_by(rotation_transform)   

    @staticmethod
    def lerp(t1, t2, progress):
        """Linear interpolation between two transforms.
            
        Parameters
        ----------
        t1 : Transform
            First transform.
        t2 : Transform
            Second transform.
        progress : float
            Progress between 0 and 1.

        Returns
        -------
        Transform
            Transform linearly interpolated between t1 and t2.
        """
        px = np.interp(progress, [0, 1], [t1.pos[0], t2.pos[0]])
        py = np.interp(progress, [0, 1], [t1.pos[1], t2.pos[1]])
        pz = np.interp(progress, [0, 1], [t1.pos[2], t2.pos[2]])
        p = np.array([px, py, pz])
        return Transform(p, quaternion.slerp(t1.orient, t2.orient, 0, 1, progress))
    
    def __eq__(self, other):
        """Equality operator for Transform objects.

        Parameters
        ----------
        other : Transform
            Transform object to compare to.

        Returns
        -------
        Bool
            True if the 3 values for positions and the 4 values for quaternions are "equal" to each other. False otherwise.  

        Note
        ----
        Equality for floating point values is defined as being within a tolerance.        
        Tolerance values are defined in the constants.py file.
        Lower tolerance values are used here for more accurate equality.
        """
        return np.allclose(self.as_matrix(), other.as_matrix(), rtol=LOW_R_TOL, atol=LOW_A_TOL)


    def __matmul__(self, other):
        """Matrix multiplication operator for Transform objects.

        Overrides the default matrix multiplication operator `@` for Transform objects.

        Parameters
        ----------
        other : Transform
            Transform object to multiply by.

        Returns
        -------
        Transform
            Resulting transform.
        """
        return self.multiply_by(other)

    def __repr__(self):
        return f'Transform({str(self.pos)}, {(self.orient)})'

identity = Transform()



class Animation:
    """Class to represent an animation. Contains a list of all joint animations for a given animation.

    Attributes:
    ----------
    joint_animations : (JointAnimation) list
        The joint animations that make up the animation.
    max_key_time : float
        The maximum time value of any key in the animation.
        Is indicative of the duration of the animation.
    """

    def __init__(self, joint_animations = [], max_key_time = 1.0):
        """**Default constructor for the Animation class.**

        Parameters:
        ----------
        joint_animations : (JointAnimation) list, optional
            The joint animations that make up the animation, by default empty list.
        max_key_time : float, optional
            The max_time of the animation. By default, 1.0.
        """
        self.joint_animations = joint_animations
        self.max_key_time = max_key_time
            
    def compute_max_key_time(self):
        """Computes the maximum time value of any key in the animation.

        Note
        ----------
        Animations loaded with the load_fbx method in skeletal_animation/utils/loader.py have a max_key_time already computed.
        """
        max_key_time = 1.0
        for joint_animation in self.joint_animations:
            max_translation_time = joint_animation.translation_curve.get_max_key_time()
            max_rotation_time = joint_animation.rotation_curve.get_max_key_time()
            max_key_time = max(max_key_time, max_translation_time, max_rotation_time)
        self.max_key_time = max_key_time

    def __repr__(self):
        return f'Animation(joint_animations={str(self.joint_animations)}, duration={str(self.duration)}'

        




from skeletal_animation.utils.loader import *

# Load the fbx file
# Replace default path with the path of the fbx file
walk_animator = load_fbx(
    "D:/He-Arc/TB/tb-animation-squelettale/skeletal_animation/animated_models/fbx/walk.fbx"
    )

# Replace default path with the path where you want to save the serialized animation
serialize(
    animator,
    "demo_hello_world",
    "D:/He-Arc/TB/tb-animation-squelettale/skeletal_animation/animated_models/serialized/"
    )



winW, winH = gh_window.getsize(0)

gx_camera.update_perspective(camera, camera_fov, 1, 0, 0, winW, winH, camera_znear, camera_zfar)

gh_camera.update_ortho(camera_ortho, -winW/2, winW/2, -winH/2, winH/2, 1.0, 10.0)
gh_camera.set_viewport(camera_ortho, 0, 0, winW, winH)

gh_mesh.update_quad_size(fullscreen_quad, winW, winH)




import unittest
import numpy as np

from skeletal_animation.core.math.constants import HIGH_R_TOL, HIGH_A_TOL
from skeletal_animation.core.animate.key import Key
from skeletal_animation.core.animate.curve import Curve

class TestCurveDefaultConstructor(unittest.TestCase):
    def setUp(self):
        self.kt0 = Key(0., [0., 0., 0.])
        self.kt1 = Key(0.33, [1., 2., 3.])
        self.kt2 = Key(20., [4., 5., 6.])

        self.kr0 = Key(0., np.quaternion(1, 0., 0., 0.))
        self.kr1 = Key(0.33, np.quaternion(2/3, 0., 2/3, 1/3))
        self.kr2 = Key(20., np.quaternion(4/3, 0., 4/3, 1/3))
    
    def test_animation_curve_default_constructor_with_default_params(self):
        curve = Curve()
        self.assertEqual(curve.keys, [])
    
    def test_animation_curve_default_constructor_with_valid_keys(self):
        with self.assertWarnsRegex(UserWarning, "Key time should be less than 1"):
            curve = Curve(keys=[self.kt0, self.kt1, self.kt2])
        self.assertEqual(curve.keys, [self.kt0, self.kt1, self.kt2])

    def test_animation_curve_default_constructor_with_duplicate_keys(self):
        kt = Key(0.33, [0., 1., 0.])
        with self.assertRaisesRegex(ValueError, "already exists in animation curve"):
            curve = Curve(keys=[self.kt0, self.kt1, kt])
            self.assertIsNone(curve)
    
    def test_animation_curve_default_constructor_with_different_key_types(self):
        with self.assertRaisesRegex(ValueError, "the type of the first key of the animation curve."):
            curve = Curve(keys=[self.kt0, self.kr1, self.kt2])
            self.assertIsNone(curve)

class TestCurveGetKeyCount(unittest.TestCase):
    def setUp(self):
        self.kt1 = Key(0.33, [1., 2., 3.])

    def test_animation_curve_get_key_count_with_no_keys(self):
        curve = Curve()
        self.assertEqual(curve.get_key_count(), 0)
    
    def test_animation_curve_get_key_count_with_key(self):
        curve = Curve(keys=[self.kt1])
        self.assertEqual(curve.get_key_count(), 1)

class TestCurveNormalizeTimes(unittest.TestCase):
    def setUp(self):
        self.kt0 = Key(0., [0., 0., 0.])
        self.kt1 = Key(0.33, [1., 2., 3.])
        self.kt2 = Key(20., [4., 5., 6.])
    
    def test_animation_curve_normalize_times_with_no_keys(self):
        curve = Curve()
        curve.normalize_times(1.)
        self.assertEqual(curve.keys, [])

    def test_animation_curve_normalize_times(self):
        with self.assertWarnsRegex(UserWarning, "Key time should be less than 1"):
            curve = Curve(keys=[self.kt0, self.kt1, self.kt2])
        curve.normalize_times(20.)
        self.assertTrue(np.allclose([key.time for key in curve.keys], [0., 0.33/20., 1.], rtol=HIGH_R_TOL, atol=HIGH_R_TOL))

class TestCurveGetKeyByTime(unittest.TestCase):
    def setUp(self):
        self.kt0 = Key(0., [0., 0., 0.])
        self.kt1 = Key(0.33, [1., 2., 3.])
        self.kt2 = Key(20., [4., 5., 6.])
        
        with self.assertWarnsRegex(UserWarning, "Key time should be less than 1"):
            self.curve0 = Curve(keys=[self.kt0, self.kt1, self.kt2])
    
    def test_animation_curve_get_key_at_with_no_keys(self):
        curve = Curve()
        self.assertIsNone(curve.get_key_at(0.))
    
    def test_animation_curve_get_key_at_with_valid_time(self):
        self.assertEqual(self.curve0.get_key_at(0.), self.kt0)
        self.assertEqual(self.curve0.get_key_at(0.33), self.kt1)
        self.assertEqual(self.curve0.get_key_at(20.), self.kt2)
    
    def test_animation_curve_get_key_at_with_invalid_time(self):
        self.assertIsNone(self.curve0.get_key_at(-0.001))
        self.assertIsNone(self.curve0.get_key_at(1/3))
        self.assertIsNone(self.curve0.get_key_at(21))
    
    def test_animation_curve_get_key_at_with_close_time(self):
        self.assertEqual(self.curve0.get_key_at(20. + 1e-5), self.kt2)


class TestCurveGetPreviousKey(unittest.TestCase):
    def setUp(self):        
        self.curve = Curve()

        for t in np.linspace(0, 1, 5):
            self.curve.add_key(Key(t, [t*1., t*2., t*3.]))

    def test_animation_curve_get_previous_key_with_no_keyss(self):
        curve = Curve()
        self.assertIsNone(curve.get_previous_key(0))
        self.assertIsNone(curve.get_previous_key(0.25))
        self.assertIsNone(curve.get_previous_key(1))

    def test_animation_curve_get_previous_key_with_keys(self):        
        self.assertEqual(self.curve.get_previous_key(-99), self.curve.keys[0])
        self.assertEqual(self.curve.get_previous_key(0), self.curve.keys[0])
        self.assertEqual(self.curve.get_previous_key(0.25), self.curve.keys[1])
        self.assertEqual(self.curve.get_previous_key(1), self.curve.keys[-1])
        self.assertEqual(self.curve.get_previous_key(99), self.curve.keys[-1])

class TestCurveGetNextKey(unittest.TestCase):
    def setUp(self):        
        self.curve = Curve()
        for t in np.linspace(0, 1, 5):
            self.curve.add_key(Key(t, [t*1., t*2., t*3.]))

    def test_animation_curve_get_next_key_with_no_keys(self):
        curve = Curve()
        self.assertIsNone(curve.get_next_key(0))
        self.assertIsNone(curve.get_next_key(0.25))
        self.assertIsNone(curve.get_next_key(1))

    def test_animation_curve_get_next_key_with_keys(self):
        self.assertEqual(self.curve.get_next_key(-99), self.curve.keys[0])
        self.assertEqual(self.curve.get_next_key(0), self.curve.keys[1])
        self.assertEqual(self.curve.get_next_key(0.25), self.curve.keys[2])
        self.assertEqual(self.curve.get_next_key(1), self.curve.keys[-1])
        self.assertEqual(self.curve.get_next_key(99), self.curve.keys[-1])
    

class TestCurveGetPreviousValue(unittest.TestCase):
    def setUp(self):        
        self.curve = Curve()
        for t in np.linspace(0, 1, 5):
            self.curve.add_key(Key(t, [t*1., t*2., t*3.]))
        
    def test_animation_curve_previous_value_with_no_keys(self):
        curve = Curve()
        self.assertIsNone(curve.get_previous_value(0))
        self.assertIsNone(curve.get_previous_value(0.25))
        self.assertIsNone(curve.get_previous_value(1))
        
    def test_animation_curve_previous_value_with_keys(self):
        self.assertTrue(np.allclose(self.curve.get_previous_value(0), [0., 0., 0.], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
        self.assertTrue(np.allclose(self.curve.get_previous_value(0.25), [0.25, 0.5, 0.75], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
        self.assertTrue(np.allclose(self.curve.get_previous_value(1), [1., 2., 3.], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
class TestCurveGetLerpValue(unittest.TestCase):
    def setUp(self):
        self.curve = Curve()
        for t in np.linspace(0, 1, 5):
            self.curve.add_key(Key(t, [t*1., t*2., t*3.]))
    
    def test_animation_curve_lerp_value_with_no_keys(self):
        curve = Curve()
        self.assertIsNone(curve.get_lerp_value(0))
        self.assertIsNone(curve.get_lerp_value(0.25))
        self.assertIsNone(curve.get_lerp_value(1))
    
    def test_animation_curve_lerp_value_with_keys(self):
        self.assertTrue(np.allclose(self.curve.get_lerp_value(0), [0., 0., 0.], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
        self.assertTrue(np.allclose(self.curve.get_lerp_value(0.1), [0.1, 0.2, 0.3], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
        self.assertTrue(np.allclose(self.curve.get_lerp_value(1), [1., 2., 3.], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))


import numpy as np

from skeletal_animation.core.math.constants import VERY_HIGH_R_TOL


def are_skeletons_similar(first_skeleton, second_skeleton):
    """Checks if the skeletons are similar.

    Similarity is defined as:
    - The skeletons have the same number of joints.
    - The local bind transforms of the joints are within a tolerance of each other.
      The tolerance used in this case does not need to be tight. Since if the joints
      of the first skeleton start at similar positions to those of the second skeleton,
      the animation will look correct.
      Tolerance values are defined in skeletal_animation/core/math/constants.py. file
      Very high tolerance values are used here.

    Parameters
    ----------
    first_skeleton : Skeleton
        First skeleton.
    second_skeleton : Skeleton
        Skeleton to compare to the first.
    
    Returns
    -------
    bool
        True if the skeletons are similar, False otherwise.

    Warnings
    --------
    This function may return false positives.
    For two skeletons to be similar, the joint hierarchy of both must be the same.
    However, this check is not implemented yet.    
    """
    first_joint_list = first_skeleton.as_joint_list()
    second_joint_list = second_skeleton.as_joint_list()

    if len(first_joint_list) != len(second_joint_list):
        return False
    for i in range(len(first_joint_list)):
        if not np.allclose(
            first_joint_list[i].local_bind_transform.as_matrix(),
            second_joint_list[i].local_bind_transform.as_matrix(),
            rtol=VERY_HIGH_R_TOL
        ):
            return False
    return True


import numpy as np

from skeletal_animation.core.math.constants import LOW_A_TOL, LOW_R_TOL

class Key:
    """Class to represent a key in an animation curve.

    It is effectively a wrapper around a time and a value.
    The time is a float representing the time of the key.
    The value can be of any type, however this class is meant to be used in Curve objects.
    Curve objects have type restrictions on the values of the keys they contain.

    Attributes:
    ----------
    time : float
        The time of the key.
    value : any
        The value of the key at a given time.
    """

    def __init__(self, time, value):
        """**Default constructor for the Key class.**

        Parameters:
        ----------
        time : float
            The time of the key.
        value : any
            The values of the key at a given time.
        """
        if not isinstance(time, (float, np.float32, np.float64)):            
            raise ValueError("The time of the key must be a float.")

        self.time = time       
        self.value = value

    def get_type(self):
        """Returns the type of the key, which is the type of its value.

        Returns:
        ----------
        type
            The type of the value of the key.
        """
        return type(self.value)

    def __repr__(self):
        return f'Key(time={str(self.time)}, value={str(self.value)})'

    def __eq__(self, other):
        """Equality operator for Key objects.

        Note
        ----
        Equality for floating point values is defined as being within a tolerance.
        Tolerance values are defined in the constants.py file.        
        Lower tolerance values are used here for more accurate equality.

        Parameters:
        ----------
        other : Key
            Key object to compare to.

        Returns:
        ----------
        bool
            True if the time and values for both keys are "equal" to each other. False otherwise.
        """
        return (np.isclose(self.time, other.time, rtol=LOW_R_TOL, atol=LOW_A_TOL)
            and self.value == other.value)